""" View for cleaner buildout calls """
from pyramid.view import view_config
from pyramid_duh import addslash

from pypicloud.route import PackagesResource
from pypicloud.views.simple import packages_to_dict


@view_config(
    context=PackagesResource,
    request_method="GET",
    subpath=(),
    renderer="packages.jinja2",
)
@addslash
def list_packages(request):
    """Render the list for all versions of all packages"""
    names = request.db.distinct()
    # remove the ones that you are not allowed to see
    names = filter(lambda x: request.access.has_permission(x, "read"), names)
    packages = []
    for package_name in names:
        packages += request.db.all(package_name)
    return {"pkgs": packages_to_dict(request, packages)}



""" Classes that provide user and package permissions """
from functools import partial

from pyramid.path import DottedNameResolver

from .aws_secrets_manager import AWSSecretsManagerAccessBackend
from .base import (
    DEFAULT_ROUNDS,
    SCHEMES,
    IAccessBackend,
    IMutableAccessBackend,
    get_pwd_context,
)
from .config import ConfigAccessBackend
from .remote import RemoteAccessBackend
from .sql import SQLAccessBackend


def includeme(config) -> None:
    """Configure the app"""
    settings = config.get_settings()

    resolver = DottedNameResolver(__name__)
    dotted_name = settings.get("pypi.auth", "config")
    if dotted_name == "config":
        dotted_name = ConfigAccessBackend
    elif dotted_name == "remote":
        dotted_name = RemoteAccessBackend
    elif dotted_name == "sql":
        dotted_name = SQLAccessBackend
    elif dotted_name == "ldap":
        dotted_name = "pypicloud.access.ldap_.LDAPAccessBackend"
    elif dotted_name == "aws_secrets_manager":
        dotted_name = AWSSecretsManagerAccessBackend
    access_backend = resolver.maybe_resolve(dotted_name)
    kwargs = access_backend.configure(settings)
    config.add_request_method(
        partial(access_backend, **kwargs), name="access", reify=True
    )
    config.add_postfork_hook(partial(access_backend.postfork, **kwargs))



""" Access backend for storing permissions in using SQLAlchemy """
import zope.sqlalchemy
from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    String,
    Table,
    Text,
    engine_from_config,
    orm,
)
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, sessionmaker

from .base import IMutableAccessBackend

# pylint: disable=C0103,W0231
Base = declarative_base()

association_table = Table(
    "pypicloud_user_groups",
    Base.metadata,
    Column(
        "username",
        String(length=255),
        ForeignKey("pypicloud_users.username", ondelete="CASCADE"),
        primary_key=True,
    ),
    Column(
        "group",
        String(length=255),
        ForeignKey("pypicloud_groups.name", ondelete="CASCADE"),
        primary_key=True,
    ),
)
# pylint: enable=C0103


class KeyVal(Base):

    """Simple model for storing key-value pairs"""

    __tablename__ = "pypicloud_keyvals"
    key = Column(String(length=255), primary_key=True)
    value = Column(Text())

    def __init__(self, key, value):
        self.key = key
        self.value = value


class User(Base):

    """User record"""

    __tablename__ = "pypicloud_users"
    username = Column(String(length=255), primary_key=True)
    password = Column("password", Text(), nullable=False)
    admin = Column(Boolean(), nullable=False)
    pending = Column(Boolean(), nullable=False)
    groups = orm.relationship(
        "Group",
        secondary=association_table,
        cascade="all",
        collection_class=set,
        backref=backref("users", collection_class=set),
    )

    def __init__(self, username, password, pending=True):
        self.username = username
        self.password = password
        self.groups = set()
        self.permissions = []
        self.admin = False
        self.pending = pending


class Group(Base):

    """Group record"""

    __tablename__ = "pypicloud_groups"
    name = Column(String(length=255), primary_key=True)

    def __init__(self, name):
        self.name = name
        self.users = set()
        self.permissions = []


class Permission(Base):

    """Base class for user and group permissions"""

    __abstract__ = True
    package = Column(String(length=255), primary_key=True)
    read = Column(Boolean())
    write = Column(Boolean())

    def __init__(self, package, read, write):
        self.package = package
        self.read = read
        self.write = write

    @property
    def permissions(self):
        """Construct permissions list"""
        perms = []
        if self.read:
            perms.append("read")
        if self.write:
            perms.append("write")
        return perms


class UserPermission(Permission):

    """Permissions for a user on a package"""

    __tablename__ = "pypicloud_user_permissions"
    username = Column(
        String(length=255),
        ForeignKey(User.username, ondelete="CASCADE"),
        primary_key=True,
    )
    user = orm.relationship(
        "User", backref=backref("permissions", cascade="all, delete-orphan")
    )

    def __init__(self, package, username, read=False, write=False):
        super(UserPermission, self).__init__(package, read, write)
        self.username = username


class GroupPermission(Permission):

    """Permissions for a group on a package"""

    __tablename__ = "pypicloud_group_permissions"
    groupname = Column(
        String(length=255), ForeignKey(Group.name, ondelete="CASCADE"), primary_key=True
    )
    group = orm.relationship(
        "Group", backref=backref("permissions", cascade="all, delete-orphan")
    )

    def __init__(self, package, groupname, read=False, write=False):
        super(GroupPermission, self).__init__(package, read, write)
        self.groupname = groupname


class SQLAccessBackend(IMutableAccessBackend):

    """
    This backend allows you to store all user and package permissions in a SQL
    database

    """

    def __init__(self, request=None, dbmaker=None, **kwargs):
        super(SQLAccessBackend, self).__init__(request, **kwargs)
        self._db = None
        self._dbmaker = dbmaker

    @property
    def db(self):
        """Lazy-create the DB session"""
        if self._db is None:
            self._db = self._dbmaker()
            if self.request is not None:
                zope.sqlalchemy.register(self._db, transaction_manager=self.request.tm)
        return self._db

    @classmethod
    def configure(cls, settings):
        kwargs = super(SQLAccessBackend, cls).configure(settings)
        engine = engine_from_config(settings, prefix="auth.db.")
        kwargs["dbmaker"] = sessionmaker(bind=engine)
        # Create SQL schema if not exists
        Base.metadata.create_all(bind=engine)
        return kwargs

    @classmethod
    def postfork(cls, **kwargs):
        # Have to dispose of connections after uWSGI forks,
        # otherwise they'll get corrupted.
        kwargs["dbmaker"].kw["bind"].dispose()

    def allow_register(self):
        ret = self.db.query(KeyVal).filter_by(key="allow_register").first()
        return ret is not None and ret.value == "true"

    def set_allow_register(self, allow):
        if allow:
            k = KeyVal("allow_register", "true")
            self.db.merge(k)
        else:
            self.db.query(KeyVal).filter_by(key="allow_register").delete()

    def _get_password_hash(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        if user:
            return user.password

    def groups(self, username=None):
        if username is None:
            query = self.db.query(Group)
            return [g.name for g in query]
        else:
            user = self.db.query(User).filter_by(username=username).first()
            if user is None:
                return []
            return [g.name for g in user.groups]

    def group_members(self, group):
        g = self.db.query(Group).filter_by(name=group).first()
        if not g:
            return []
        return [u.username for u in g.users]

    def is_admin(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        return user and user.admin

    def group_permissions(self, package):
        query = self.db.query(GroupPermission).filter_by(package=package)
        perms = {}

        for perm in query:
            perms[perm.groupname] = perm.permissions
        return perms

    def user_permissions(self, package):
        query = self.db.query(UserPermission).filter_by(package=package)
        perms = {}
        for perm in query:
            perms[perm.username] = perm.permissions
        return perms

    def user_package_permissions(self, username):
        query = self.db.query(UserPermission).filter_by(username=username)
        packages = []
        for perm in query:
            packages.append({"package": perm.package, "permissions": perm.permissions})
        return packages

    def group_package_permissions(self, group):
        query = self.db.query(GroupPermission).filter_by(groupname=group)
        packages = []
        for perm in query:
            packages.append({"package": perm.package, "permissions": perm.permissions})
        return packages

    def user_data(self, username=None):
        if username is None:
            query = self.db.query(User).filter_by(pending=False)
            users = []
            for user in query:
                users.append({"username": user.username, "admin": user.admin})
            return users
        else:
            user = (
                self.db.query(User).filter_by(username=username, pending=False).first()
            )
            if user is not None:
                return {
                    "username": user.username,
                    "admin": user.admin,
                    "groups": [g.name for g in user.groups],
                }

    def need_admin(self):
        return self.db.query(User).filter_by(admin=True).first() is None

    def _register(self, username, password):
        user = User(username, password)
        self.db.add(user)

    def pending_users(self):
        query = self.db.query(User).filter_by(pending=True)
        return [u.username for u in query]

    def approve_user(self, username):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.pending = False

    def _set_password_hash(self, username, password_hash):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.password = password_hash

    def delete_user(self, username):
        self.db.query(User).filter_by(username=username).delete()
        clause = association_table.c.username == username
        self.db.execute(association_table.delete(clause))

    def set_user_admin(self, username, admin):
        user = self.db.query(User).filter_by(username=username).first()
        if user is not None:
            user.admin = admin

    def edit_user_group(self, username, groupname, add):
        user = self.db.query(User).filter_by(username=username).first()
        group = self.db.query(Group).filter_by(name=groupname).first()
        if user is not None and group is not None:
            if add:
                user.groups.add(group)
            else:
                user.groups.remove(group)

    def create_group(self, group):
        self.db.add(Group(group))

    def delete_group(self, group):
        self.db.query(Group).filter_by(name=group).delete()
        clause = association_table.c.group == group
        self.db.execute(association_table.delete(clause))

    def edit_user_permission(self, package, username, perm, add):
        record = (
            self.db.query(UserPermission)
            .filter_by(package=package, username=username)
            .first()
        )
        if record is None:
            if not add:
                return
            record = UserPermission(package, username)
            self.db.add(record)
        if perm == "read":
            record.read = add
        elif perm == "write":
            record.write = add
        else:
            raise ValueError("Unrecognized permission '%s'" % perm)
        if not record.read and not record.write:
            self.db.delete(record)

    def edit_group_permission(self, package, group, perm, add):
        record = (
            self.db.query(GroupPermission)
            .filter_by(package=package, groupname=group)
            .first()
        )
        if record is None:
            if not add:
                return
            record = GroupPermission(package, group)
            self.db.add(record)
        if perm == "read":
            record.read = add
        elif perm == "write":
            record.write = add
        else:
            raise ValueError("Unrecognized permission '%s'" % perm)
        if not record.read and not record.write:
            self.db.delete(record)

    def check_health(self):
        try:
            self.db.query(KeyVal).first()
        except SQLAlchemyError as e:
            return (False, str(e))
        else:
            return (True, "")



""" Base class for storage backends """
from typing import BinaryIO, List, Tuple, Type

from pyramid.request import Request

from pypicloud.models import Package


class IStorage(object):

    """Base class for a backend that stores package files"""

    def __init__(self, request: Request):
        self.request = request

    @classmethod
    def configure(cls, settings):
        """Configure the storage method with app settings"""
        return {}

    def list(self, factory: Type[Package] = Package) -> List[Package]:
        """Return a list or generator of all packages"""
        raise NotImplementedError

    def get_url(self, package: Package) -> str:
        """
        Create or return an HTTP url for a package file

        By default this will return a link to the download endpoint

        /api/package/<package>/<filename>

        Returns
        -------
        link : str
            Link to the location of this package file

        """
        fragment = ""
        if package.data.get("hash_sha256"):
            fragment = "sha256=" + package.data["hash_sha256"]
        return self.request.app_url(
            "api", "package", package.name, package.filename, fragment=fragment
        )

    def download_response(self, package: Package):
        """
        Return a HTTP Response that will download this package

        This is called from the download endpoint

        """
        raise NotImplementedError

    def upload(self, package: Package, datastream: BinaryIO) -> None:
        """
        Upload a package file to the storage backend

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`
            The package metadata
        datastream : file
            A file-like object that contains the package data

        """
        raise NotImplementedError

    def delete(self, package: Package) -> None:
        """
        Delete a package file

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`
            The package metadata

        """
        raise NotImplementedError

    def open(self, package: Package):
        """
        Get a buffer object that can read the package data

        This should be a context manager. It is used in migration scripts, not
        directly by the web application.

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        Examples
        --------
        ::

            with storage.open(package) as pkg_data:
                with open('outfile.tar.gz', 'w') as ofile:
                    ofile.write(pkg_data.read())

        """
        raise NotImplementedError

    def check_health(self) -> Tuple[bool, str]:
        """
        Check the health of the storage backend

        Returns
        -------
        (healthy, status) : (bool, str)
            Tuple that describes the health status and provides an optional
            status message

        """
        return True, ""



""" Store packages in GCS """
import io
import json
import logging
import os
import posixpath
from datetime import timedelta

from google.auth import compute_engine
from google.auth.transport import requests
from google.cloud import storage
from pyramid.settings import asbool

from pypicloud.models import Package

from .object_store import ObjectStoreStorage

LOG = logging.getLogger(__name__)


class GoogleCloudStorage(ObjectStoreStorage):

    """Storage backend that uses GCS"""

    test = False

    def __init__(
        self,
        request=None,
        bucket_factory=None,
        service_account_json_filename=None,
        project_id=None,
        use_iam_signer=False,
        iam_signer_service_account_email=None,
        **kwargs
    ):
        super(GoogleCloudStorage, self).__init__(request=request, **kwargs)

        self._bucket = None
        self._bucket_factory = bucket_factory
        self.use_iam_signer = use_iam_signer
        self.iam_signer_service_account_email = iam_signer_service_account_email

        if self.public_url:
            raise NotImplementedError(
                "GoogleCloudStorage backend does not yet support public URLs"
            )

        if self.sse:
            raise NotImplementedError(
                "GoogleCloudStorage backend does not yet support customized "
                "server-side encryption"
            )

    @classmethod
    def _subclass_specific_config(cls, settings, common_config):
        """Extract GCP-specific config settings: specifically, the path to
        the service account key file, and the project id.  Both are
        optional.
        """
        service_account_json_filename = settings.get(
            "storage.gcp_service_account_json_filename"
        ) or os.getenv("GOOGLE_APPLICATION_CREDENTIALS")

        if (
            service_account_json_filename
            and not os.path.isfile(service_account_json_filename)
            and not cls.test
        ):
            raise Exception(
                "Service account json file not found at path {}".format(
                    service_account_json_filename
                )
            )

        bucket_name = settings.get("storage.bucket")
        if bucket_name is None:
            raise ValueError("You must specify the 'storage.bucket'")

        iam_signer_service_account_email = settings.get(
            "storage.iam_signer_service_account_email"
        )
        if iam_signer_service_account_email is None and service_account_json_filename:
            with io.open(service_account_json_filename, "r", encoding="utf-8") as ifile:
                credentials = json.load(ifile)
            iam_signer_service_account_email = credentials.get("client_email")

        return {
            "service_account_json_filename": service_account_json_filename,
            "project_id": settings.get("storage.gcp_project_id"),
            "use_iam_signer": asbool(settings.get("storage.gcp_use_iam_signer", False)),
            "iam_signer_service_account_email": iam_signer_service_account_email,
            "bucket_factory": lambda: cls.get_bucket(bucket_name, settings),
        }

    @classmethod
    def _get_storage_client(cls, settings):
        """Helper method for constructing a properly-configured GCS client
        object from the provided settings.
        """
        client_settings = cls._subclass_specific_config(settings, {})

        client_args = {}
        if client_settings["project_id"]:
            LOG.info("Using GCP project id `%s`", client_settings["project_id"])
            client_args["project"] = client_settings["project_id"]

        service_account_json_filename = client_settings.get(
            "service_account_json_filename"
        )

        if not service_account_json_filename:
            LOG.info("Creating GCS client without service account JSON file")

            client = storage.Client(**client_args)
        else:
            if not os.path.isfile(service_account_json_filename) and not cls.test:
                raise Exception(
                    "Service account JSON file not found at provided "
                    "path {}".format(service_account_json_filename)
                )

            LOG.info(
                "Creating GCS client from service account JSON file %s",
                service_account_json_filename,
            )

            client = storage.Client.from_service_account_json(
                service_account_json_filename, **client_args
            )

        return client

    @classmethod
    def get_bucket(cls, bucket_name, settings):
        client = cls._get_storage_client(settings)

        bucket = client.bucket(bucket_name)

        if not bucket.exists():
            bucket.location = settings.get("storage.region_name")
            LOG.info(
                "Creating GCS bucket %s in location %s", bucket_name, bucket.location
            )
            bucket.create()

        return bucket

    @classmethod
    def package_from_object(cls, blob, factory):
        """Create a package from a GCS object"""
        filename = posixpath.basename(blob.name)
        if blob.metadata is None:
            return None
        name = blob.metadata.get("name")
        version = blob.metadata.get("version")
        if name is None or version is None:
            return None
        metadata = Package.read_metadata(blob.metadata)
        return factory(
            name, version, filename, blob.updated, path=blob.name, **metadata
        )

    @property
    def bucket(self):
        if self._bucket is None:
            self._bucket = self._bucket_factory()
        return self._bucket

    def list(self, factory=Package):
        blobs = self.bucket.list_blobs(prefix=self.bucket_prefix or None)
        for blob in blobs:
            pkg = self.package_from_object(blob, factory)
            if pkg is not None:
                yield pkg

    def _generate_url(self, package):
        """Generate a signed url to the GCS file"""
        blob = self._get_gcs_blob(package)

        if self.use_iam_signer:
            # Workaround for https://github.com/googleapis/google-auth-library-python/issues/50
            signing_credentials = compute_engine.IDTokenCredentials(
                requests.Request(),
                "",
                service_account_email=self.iam_signer_service_account_email,
            )
        else:
            signing_credentials = None
        return blob.generate_signed_url(
            expiration=timedelta(seconds=self.expire_after),
            credentials=signing_credentials,
            version="v4",
        )

    def _get_gcs_blob(self, package):
        """Get a GCS blob object for the specified package"""
        return self.bucket.blob(self.get_path(package))

    def upload(self, package, datastream):
        """Upload the package to GCS"""
        metadata = {"name": package.name, "version": package.version}
        metadata.update(package.get_metadata())

        blob = self._get_gcs_blob(package)

        blob.metadata = metadata

        blob.upload_from_file(datastream, predefined_acl=self.object_acl)

        if self.storage_class is not None:
            blob.update_storage_class(self.storage_class)

    def delete(self, package):
        """Delete the package"""
        blob = self._get_gcs_blob(package)
        blob.delete()



""" Tools and resources for traversal routing """
import functools
from typing import Any, Callable, Dict


class IStaticResource(object):

    """Simple resource base class for static-mapping of paths"""

    __name__ = ""
    __parent__ = None
    subobjects = {}  # type: Dict[str, Callable[[Any], Any]]

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = self.subobjects[name](self.request)
        child.__parent__ = self
        child.__name__ = name
        return child


class IResourceFactory(object):

    """Resource that generates child resources from a factory"""

    __name__ = ""
    __parent__ = None

    def __factory__(self, name):
        return None

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = self.__factory__(name)  # pylint: disable=E1128
        child.__name__ = name
        child.__parent__ = self
        return child


class SimpleResource(object):

    """Resource for simple pip calls"""

    def __init__(self, request):
        self.request = request

    def __getitem__(self, name):
        child = SimplePackageResource(self.request, name)
        child.__parent__ = self
        child.__name__ = name
        return child


class SimplePackageResource(object):

    """Resource for requesting simple endpoint package versions"""

    __parent__ = None
    __name__ = None

    def __init__(self, request, name):
        self.request = request
        self.name = name
        self.__acl__ = request.access.get_acl(self.name)


class APIPackagingResource(IResourceFactory):

    """Resource for api package queries"""

    def __init__(self, request):
        super(APIPackagingResource, self).__init__(request)
        self.__factory__ = functools.partial(APIPackageResource, self.request)


class APIPackageResource(IResourceFactory):

    """Resource for requesting package versions"""

    def __init__(self, request, name):
        super(APIPackageResource, self).__init__(request)
        self.name = name
        self.__factory__ = functools.partial(
            APIPackageFileResource, self.request, self.name
        )
        self.__acl__ = request.access.get_acl(self.name)


class APIPackageFileResource(object):

    """Resource for api endpoints dealing with a single package version"""

    __parent__ = None
    __name__ = None

    def __init__(self, request, name, filename):
        self.request = request
        self.name = name
        self.filename = filename


class APIResource(IStaticResource):

    """Resource for api calls"""

    subobjects = {"package": APIPackagingResource}


class AdminResource(IStaticResource):

    """Resource for admin calls"""


class PackagesResource(IStaticResource):

    """Resource for cleaner buildout config"""


class Root(IStaticResource):

    """Root context for PyPI Cloud"""

    subobjects = {
        "api": APIResource,
        "admin": AdminResource,
        "simple": SimpleResource,
        "pypi": SimpleResource,
        "packages": PackagesResource,
    }

    def __init__(self, request):
        super(Root, self).__init__(request)
        self.__acl__ = request.access.ROOT_ACL



""" Storage backend implementations """
from functools import partial
from typing import Any, Callable

from pyramid.path import DottedNameResolver

from .base import IStorage
from .files import FileStorage
from .s3 import CloudFrontS3Storage, S3Storage

try:
    from .gcs import GoogleCloudStorage

    GCS_IS_AVAILABLE = True
except ImportError:
    GCS_IS_AVAILABLE = False

try:
    from .azure_blob import AzureBlobStorage

    AZURE_BLOB_IS_AVAILABLE = True
except ImportError:
    AZURE_BLOB_IS_AVAILABLE = False


def get_storage_impl(settings) -> Callable[[Any], Any]:
    """Get and configure the storage backend wrapper"""
    resolver = DottedNameResolver(__name__)
    storage = settings.get("pypi.storage", "file")
    if storage == "azure-blob":
        if not AZURE_BLOB_IS_AVAILABLE:
            raise ValueError(
                "azure-blob storage backend selected but Azure Blob "
                "Storage is not available. "
                "Please install the azure-storage-blob library by "
                "including the `azure-blob` extra in your pip-install step. "
                "For example: `pip install pypicloud[azure-blob]`"
            )

        storage = "pypicloud.storage.AzureBlobStorage"
    elif storage == "s3":
        storage = "pypicloud.storage.S3Storage"
    elif storage == "cloudfront":
        storage = "pypicloud.storage.CloudFrontS3Storage"
    elif storage == "gcs":
        if not GCS_IS_AVAILABLE:
            raise ValueError(
                "gcs backend selected but GCS is not available. "
                "Please install the google-cloud-storage library by "
                "including the `gcs` extra in your pip-install step. "
                "For example: `pip install pypicloud[gcs]`"
            )

        storage = "pypicloud.storage.GoogleCloudStorage"
    elif storage == "file":
        storage = "pypicloud.storage.FileStorage"
    storage_impl = resolver.resolve(storage)
    kwargs = storage_impl.configure(settings)
    return partial(storage_impl, **kwargs)



""" Backend that defers to another server for access control """
from .base import IAccessBackend


class RemoteAccessBackend(IAccessBackend):

    """
    This backend allows you to defer all user auth and permissions to a remote
    server. It requires the ``requests`` package.

    """

    def __init__(self, request=None, settings=None, server=None, auth=None, **kwargs):
        super(RemoteAccessBackend, self).__init__(request, **kwargs)
        self._settings = settings
        self.server = server
        self.auth = auth

    @classmethod
    def configure(cls, settings):
        kwargs = super(RemoteAccessBackend, cls).configure(settings)
        kwargs["settings"] = settings
        kwargs["server"] = settings["auth.backend_server"]
        auth = None
        user = settings.get("auth.user")
        if user is not None:
            password = settings.get("auth.password")
            auth = (user, password)
        kwargs["auth"] = auth
        return kwargs

    def _req(self, uri, params=None):
        """Hit a server endpoint and return the json response"""
        try:
            import requests
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "You must 'pip install requests' before using "
                "the remote server access backend"
            ) from e
        response = requests.get(self.server + uri, params=params, auth=self.auth)
        response.raise_for_status()
        return response.json()

    def verify_user(self, username, password):
        uri = self._settings.get("auth.uri.verify", "/verify")
        params = {"username": username, "password": password}
        return self._req(uri, params)

    def _get_password_hash(self, username):
        # We don't have to do anything here because we overrode 'verify_user'
        pass

    def groups(self, username=None):
        uri = self._settings.get("auth.uri.groups", "/groups")
        params = {}
        if username is not None:
            params["username"] = username
        return self._req(uri, params)

    def group_members(self, group):
        uri = self._settings.get("auth.uri.group_members", "/group_members")
        params = {"group": group}
        return self._req(uri, params)

    def is_admin(self, username):
        uri = self._settings.get("auth.uri.admin", "/admin")
        params = {"username": username}
        return self._req(uri, params)

    def group_permissions(self, package):
        uri = self._settings.get("auth.uri.group_permissions", "/group_permissions")
        params = {"package": package}
        return self._req(uri, params)

    def user_permissions(self, package):
        uri = self._settings.get("auth.uri.user_permissions", "/user_permissions")
        params = {"package": package}
        return self._req(uri, params)

    def user_package_permissions(self, username):
        uri = self._settings.get(
            "auth.uri.user_package_permissions", "/user_package_permissions"
        )
        params = {"username": username}
        return self._req(uri, params)

    def group_package_permissions(self, group):
        uri = self._settings.get(
            "auth.uri.group_package_permissions", "/group_package_permissions"
        )
        params = {"group": group}
        return self._req(uri, params)

    def user_data(self, username=None):
        uri = self._settings.get("auth.uri.user_data", "/user_data")
        params = None
        if username is not None:
            params = {"username": username}
        return self._req(uri, params)



""" Utilities """
import logging
import os
import re
import time
import unicodedata
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from distlib.locators import Locator
from distlib.util import split_filename
from distlib.wheel import Wheel

LOG = logging.getLogger(__name__)
ALL_EXTENSIONS = Locator.source_extensions + Locator.binary_extensions
SENTINEL = object()


def parse_filename(filename: str, name: Optional[str] = None) -> Tuple[str, str]:
    """Parse a name and version out of a filename"""
    version = None
    for ext in ALL_EXTENSIONS:
        if filename.endswith(ext):
            if ext == ".whl":
                wheel = Wheel(filename)
                return wheel.name, wheel.version
            trimmed = filename[: -len(ext)]
            parsed = split_filename(trimmed, name)
            if parsed is None:
                break
            else:
                parsed_name, version = parsed[:2]
            break
    if version is None:
        raise ValueError("Cannot parse package file '%s'" % filename)
    if name is None:
        name = parsed_name
    return normalize_name(name), version


def get_packagetype(name: str) -> str:
    """Get package type out of a filename"""
    if name.endswith(".tar.gz"):
        return "sdist"
    elif name.endswith(".egg"):
        return "bdist_egg"
    elif name.endswith(".whl"):
        return "bdist_wheel"
    else:
        return ""


def normalize_name(name: str) -> str:
    """Normalize a python package name"""
    # Lifted directly from PEP503:
    # https://www.python.org/dev/peps/pep-0503/#id4
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_metadata_value(value: Union[str, bytes]) -> str:
    """Strip non-ASCII characters from metadata values"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    if isinstance(value, str):
        value = "".join(c for c in unicodedata.normalize("NFKD", value) if ord(c) < 128)
    return value


def normalize_metadata(metadata: Dict[str, Union[str, bytes]]) -> Dict[str, str]:
    """
    Strip non-ASCII characters from metadata values
    and replace "_" in metadata keys to "-"
    """
    return {
        key.replace("_", "-"): normalize_metadata_value(value)
        for key, value in metadata.items()
    }


def create_matcher(queries: List[str], query_type: str) -> Callable[[str], bool]:
    """
    Create a matcher for a list of queries

    Parameters
    ----------
    queries : list
        List of queries

    query_type: str
        Type of query to run: ["or"|"and"]

    Returns
    -------
        Matcher function

    """
    queries = [query.lower() for query in queries]
    if query_type == "or":
        return lambda x: any((q in x.lower() for q in queries))
    else:
        return lambda x: all((q in x.lower() for q in queries))


def get_environ_setting(settings: dict, key: str, default=None):
    env_key = "PPC_" + key.upper().replace(".", "_")
    return os.environ.get(env_key, settings.get(key, default))


def get_settings(settings: dict, prefix: str, **kwargs) -> dict:
    """
    Convenience method for fetching settings

    Returns a dict; any settings that were missing from the config file will
    not be present in the returned dict (as opposed to being present with a
    None value)

    Parameters
    ----------
    settings : dict
        The settings dict
    prefix : str
        String to prefix all keys with when fetching value from settings
    **kwargs : dict
        Mapping of setting name to conversion function (e.g. str or asbool)

    """
    computed = {}
    for name, fxn in kwargs.items():
        val = settings.get(prefix + name)
        if val is not None:
            computed[name] = fxn(val)
    return computed


class TimedCache(dict):
    """
    Dict that will store entries for a given time, then evict them

    Parameters
    ----------
    cache_time : int or None
        The amount of time to cache entries for, in seconds. 0 will not cache.
        None will cache forever.
    factory : callable, optional
        If provided, when the TimedCache is accessed and has no value, it will
        attempt to populate itself by calling this function with the key it was
        accessed with. This function should return a value to cache, or None if
        no value is found.

    """

    def __init__(
        self, cache_time: Optional[int], factory: Optional[Callable[[Any], Any]] = None
    ):
        super(TimedCache, self).__init__()
        if cache_time is not None and cache_time < 0:
            raise ValueError("cache_time cannot be negative")
        self._cache_time = cache_time
        self._factory = factory
        self._times = {}  # type: Dict[str, float]

    def _has_expired(self, key):
        """Check if a key is both present and expired"""
        if key not in self._times or self._cache_time is None:
            return False
        updated = self._times[key]
        return updated is not None and time.time() - updated > self._cache_time

    def _evict(self, key):
        """Remove a key if it has expired"""
        if self._has_expired(key):
            del self[key]

    def __contains__(self, key):
        self._evict(key)
        return super(TimedCache, self).__contains__(key)

    def __delitem__(self, key):
        del self._times[key]
        super(TimedCache, self).__delitem__(key)

    def __setitem__(self, key, value):
        if self._cache_time == 0:
            return
        self._times[key] = time.time()
        super(TimedCache, self).__setitem__(key, value)

    def __getitem__(self, key):
        self._evict(key)
        try:
            value = super(TimedCache, self).__getitem__(key)
        except KeyError:
            if self._factory:
                value = self._factory(key)
                if value is None:
                    raise
                self[key] = value
            else:
                raise
        return value

    def get(self, key, default=None):
        self._evict(key)
        value = super(TimedCache, self).get(key, SENTINEL)
        if value is SENTINEL:
            if self._factory is not None:
                value = self._factory(key)
                if value is not None:
                    self[key] = value
                    return value
                else:
                    return default
            else:
                return default
        else:
            return value

    def set_expire(self, key, value, expiration):
        """
        Set a value in the cache with a specific expiration

        Parameters
        ----------
        key : str
        value : value
        expiration : int or None
            Sets the value to expire this many seconds from now. If None, will
            never expire.

        """
        if expiration is not None:
            if expiration <= 0:
                try:
                    del self[key]
                except KeyError:
                    pass
                return
            expiration = time.time() + expiration - self._cache_time

        self._times[key] = expiration
        super(TimedCache, self).__setitem__(key, value)



""" Store packages in Azure Blob Storage """
import logging
import os
import posixpath
from contextlib import contextmanager
from datetime import timedelta
from io import BytesIO
from urllib.request import urlopen

from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import BlobSasPermissions, BlobServiceClient, generate_blob_sas
from pyramid.httpexceptions import HTTPFound
from pyramid.settings import asbool

from pypicloud.dateutil import utcnow
from pypicloud.models import Package

from .base import IStorage

LOG = logging.getLogger(__name__)


class AzureBlobStorage(IStorage):
    """Storage backend that uses Azure Blob Storage"""

    test = False

    def __init__(
        self,
        request,
        expire_after=None,
        path_prefix=None,
        redirect_urls=None,
        storage_account_name=None,
        storage_account_key=None,
        storage_container_name=None,
    ):
        super(AzureBlobStorage, self).__init__(request)

        self.expire_after = expire_after
        self.path_prefix = path_prefix
        self.redirect_urls = redirect_urls
        self.storage_account_name = storage_account_name
        self.storage_account_key = storage_account_key
        self.storage_container_name = storage_container_name
        self.azure_storage_account_url = "https://{}.blob.core.windows.net".format(
            storage_account_name
        )
        self.blob_service_client = BlobServiceClient(
            account_url=self.azure_storage_account_url,
            credential=self.storage_account_key,
        )
        self.container_client = self.blob_service_client.get_container_client(
            self.storage_container_name
        )

    @classmethod
    def configure(cls, settings):
        def get_setting(key, env_key, default=None):
            if key in settings:
                return settings[key]
            else:
                return os.environ.get(env_key, default)

        kwargs = super(AzureBlobStorage, cls).configure(settings)
        kwargs["expire_after"] = int(settings.get("storage.expire_after", 60 * 60 * 24))
        kwargs["path_prefix"] = settings.get("storage.prefix", "")
        kwargs["redirect_urls"] = asbool(settings.get("storage.redirect_urls", True))
        kwargs["storage_account_name"] = get_setting(
            "storage.storage_account_name", "AZURE_STORAGE_ACCOUNT"
        )
        if kwargs["storage_account_name"] is None:
            raise ValueError("You must specify the 'storage.storage_account_name'")

        kwargs["storage_account_key"] = get_setting(
            "storage.storage_account_key", "AZURE_STORAGE_KEY"
        )
        if kwargs["storage_account_key"] is None:
            raise ValueError("You must specify the 'storage.storage_account_key'")

        kwargs["storage_container_name"] = settings.get(
            "storage.storage_container_name"
        )
        if kwargs["storage_container_name"] is None:
            raise ValueError("You must specify the 'storage.storage_container_name'")

        return kwargs

    def _generate_url(self, package: Package) -> str:
        path = self.get_path(package)

        url_params = generate_blob_sas(
            account_name=self.storage_account_name,
            container_name=self.storage_container_name,
            blob_name=path,
            account_key=self.storage_account_key,
            permission=BlobSasPermissions(read=True),
            expiry=utcnow() + timedelta(seconds=self.expire_after),
            protocol="https",
        )

        url = "{}/{}/{}?{}".format(
            self.azure_storage_account_url,
            self.storage_container_name,
            path,
            url_params,
        )
        return url

    def download_response(self, package):
        return HTTPFound(location=self._generate_url(package))

    def list(self, factory=Package):
        # List does not return metadata :(
        for blob_properties in self.container_client.list_blobs(
            name_starts_with=self.path_prefix
        ):
            blob_client = self.container_client.get_blob_client(
                blob=blob_properties.name
            )
            metadata = blob_client.get_blob_properties()

            yield factory(
                metadata.metadata["name"],
                metadata.metadata["version"],
                posixpath.basename(blob_properties.name),
                blob_properties.last_modified,
                path=blob_properties.name,
                **Package.read_metadata(metadata.metadata)
            )

    def get_path(self, package):
        """Get the fully-qualified bucket path for a package"""
        if "path" not in package.data:
            package.data["path"] = (
                self.path_prefix + package.name + "/" + package.filename
            )
        return package.data["path"]

    def upload(self, package, datastream):
        path = self.get_path(package)

        metadata = package.get_metadata()
        metadata["name"] = package.name
        metadata["version"] = package.version

        blob_client = self.container_client.get_blob_client(blob=path)
        blob_client.upload_blob(data=datastream, metadata=metadata)

    def delete(self, package):
        path = self.get_path(package)
        blob_client = self.container_client.get_blob_client(blob=path)
        blob_client.delete_blob()

    def check_health(self):
        try:
            self.container_client.get_blob_client(
                blob="__notexist"
            ).get_blob_properties()
        except ResourceNotFoundError:
            pass
        except Exception as e:
            return False, str(e)
        return True, ""

    @contextmanager
    def open(self, package):
        url = self._generate_url(package)
        handle = urlopen(url)
        try:
            yield BytesIO(handle.read())
        finally:
            handle.close()



""" Utilities for authentication and authorization """
import binascii
from base64 import b64decode

# pylint: disable=E0611
from paste.httpheaders import AUTHORIZATION, WWW_AUTHENTICATE

# pylint: enable=E0611
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPForbidden, HTTPUnauthorized
from pyramid.interfaces import ISecurityPolicy
from pyramid.security import Everyone
from zope.interface import implementer


# Copied from
# http://docs.pylonsproject.org/projects/pyramid_cookbook/en/latest/auth/basic.html
def get_basicauth_credentials(request):
    """Get the user/password from HTTP basic auth"""
    authorization = AUTHORIZATION(request.environ)
    try:
        authmeth, auth = authorization.split(" ", 1)
    except ValueError:  # not enough values to unpack
        return None
    if authmeth.lower() == "basic":
        try:
            auth = b64decode(auth.strip()).decode("utf8")
        except (TypeError, binascii.Error):  # can't decode
            return None
        try:
            login, password = auth.split(":", 1)
        except ValueError:  # not enough values to unpack
            return None
        return {"login": login, "password": password}

    return None


@implementer(ISecurityPolicy)
class PypicloudSecurityPolicy:
    def __init__(self):
        self.acl_policy = ACLAuthorizationPolicy()

    def identity(self, request):
        """Return the :term:`identity` of the current user.  The object can be
        of any shape, such as a simple ID string or an ORM object.
        """
        # First try fetching from the session
        userid = request.session.get("user", None)
        if userid is not None:
            return userid
        # Then fall back to HTTP basic auth
        credentials = get_basicauth_credentials(request)
        if credentials is None:
            return None
        userid = credentials["login"]
        if request.access.verify_user(userid, credentials["password"]):
            return userid
        return None

    def authenticated_userid(self, request):
        """Return a :term:`userid` string identifying the trusted and
        verified user, or ``None`` if unauthenticated.

        If the result is ``None``, then
        :attr:`pyramid.request.Request.is_authenticated` will return ``False``.
        """
        return self.identity(request)

    def effective_principals(self, request):
        userid = self.authenticated_userid(request)
        if userid is None:
            return [Everyone]
        return request.access.user_principals(userid)

    def permits(self, request, context, permission):
        """Return an instance of :class:`pyramid.security.Allowed` if a user
        of the given identity is allowed the ``permission`` in the current
        ``context``, else return an instance of
        :class:`pyramid.security.Denied`.
        """
        principals = self.effective_principals(request)
        return self.acl_policy.permits(context, principals, permission)

    def remember(self, request, userid, **kw):
        """Return a set of headers suitable for 'remembering' the
        :term:`userid` named ``userid`` when set in a response.  An individual
        security policy and its consumers can decide on the composition and
        meaning of ``**kw``.
        """
        request.session["user"] = userid
        return []

    def forget(self, request, **kw):
        """Return a set of headers suitable for 'forgetting' the
        current user on subsequent requests.  An individual security policy and
        its consumers can decide on the composition and meaning of ``**kw``.
        """
        request.session.delete()
        return []


def _request_login(request):
    """Return a 401 to force pip to upload its HTTP basic auth credentials"""
    response = HTTPUnauthorized()
    realm = WWW_AUTHENTICATE.tuples('Basic realm="%s"' % request.registry.realm)
    response.headers.update(realm)
    return response


def _forbid(request):
    """
    Return a 403 if user is logged in, otherwise return a 401.

    This is required to force pip to upload its HTTP basic auth credentials

    """
    if request.is_authenticated:
        return HTTPForbidden()
    else:
        return _request_login(request)


def includeme(config):
    """Configure the app"""
    config.set_security_policy(PypicloudSecurityPolicy())
    config.add_request_method(_forbid, name="forbid")
    config.add_request_method(_request_login, name="request_login")

    settings = config.get_settings()
    config.registry.realm = settings.get("pypi.realm", "pypi")



""" Views for simple api calls that return json data """
import logging
import posixpath
from contextlib import closing
from io import BytesIO
from urllib.request import urlopen

# pylint: disable=E0611
from paste.httpheaders import CACHE_CONTROL, CONTENT_DISPOSITION

# pylint: enable=E0611
from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden, HTTPNotFound
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.view import view_config
from pyramid_duh import addslash, argify

from pypicloud.route import (
    APIPackageFileResource,
    APIPackageResource,
    APIPackagingResource,
    APIResource,
)
from pypicloud.util import normalize_name

from .login import handle_register_request

LOG = logging.getLogger(__name__)


@view_config(
    context=APIPackagingResource, request_method="GET", subpath=(), renderer="json"
)
@addslash
@argify
def all_packages(request, verbose=False):
    """List all packages"""
    if verbose:
        packages = request.db.summary()
    else:
        packages = request.db.distinct()
    i = 0
    while i < len(packages):
        package = packages[i]
        name = package if isinstance(package, str) else package["name"]
        if not request.access.has_permission(name, "read"):
            del packages[i]
            continue
        i += 1
    return {"packages": packages}


@view_config(
    context=APIPackageResource,
    request_method="GET",
    subpath=(),
    renderer="json",
    permission="read",
)
@addslash
def package_versions(context, request):
    """List all unique package versions"""
    normalized_name = normalize_name(context.name)
    versions = request.db.all(normalized_name)
    return {
        "packages": versions,
        "write": request.access.has_permission(normalized_name, "write"),
    }


def fetch_dist(request, url, name, version, summary, requires_python):
    """Fetch a Distribution and upload it to the storage backend"""
    filename = posixpath.basename(url)
    handle = urlopen(url)
    with closing(handle):
        data = handle.read()
    # TODO: digest validation
    return (
        request.db.upload(
            filename, BytesIO(data), name, version, summary, requires_python
        ),
        data,
    )


@view_config(context=APIPackageFileResource, request_method="GET", permission="read")
def download_package(context, request):
    """Download package, or redirect to the download link"""
    package = request.db.fetch(context.filename)
    if not package:
        if request.registry.fallback != "cache":
            return HTTPNotFound()
        if not request.access.can_update_cache():
            return request.forbid()
        # If we are caching pypi, download the package from pypi and save it
        releases = request.locator.get_releases(context.name)

        dist = None
        for release in releases:
            if posixpath.basename(release["url"]) == context.filename:
                dist = release
                break
        if dist is None:
            return HTTPNotFound()
        LOG.info("Caching %s from %s", context.filename, request.fallback_simple)
        package, data = fetch_dist(
            request,
            dist["url"],
            dist["name"],
            dist["version"],
            dist["summary"],
            dist["requires_python"],
        )
        disp = CONTENT_DISPOSITION.tuples(filename=package.filename)
        request.response.headers.update(disp)
        cache_control = CACHE_CONTROL.tuples(
            public=True, max_age=request.registry.package_max_age
        )
        request.response.headers.update(cache_control)
        request.response.body = data
        request.response.content_type = "application/octet-stream"
        return request.response
    if request.registry.stream_files:
        with request.db.storage.open(package) as data:
            request.response.body = data.read()
        disp = CONTENT_DISPOSITION.tuples(filename=package.filename)
        request.response.headers.update(disp)
        cache = CACHE_CONTROL.tuples(
            public=True, max_age=request.registry.package_max_age
        )
        request.response.headers.update(cache)
        request.response.content_type = "application/octet-stream"
        return request.response
    response = request.db.download_response(package)
    return response


@view_config(
    context=APIPackageFileResource,
    request_method="POST",
    subpath=(),
    renderer="json",
    permission="write",
)
@argify
def upload_package(context, request, content, summary=None, requires_python=None):
    """Upload a package"""
    try:
        return request.db.upload(
            content.filename,
            content.file,
            name=context.name,
            summary=summary,
            requires_python=requires_python,
        )
    except ValueError as e:  # pragma: no cover
        return HTTPBadRequest(*e.args)


@view_config(
    context=APIPackageFileResource,
    request_method="DELETE",
    subpath=(),
    permission="write",
)
def delete_package(context, request):
    """Delete a package"""
    package = request.db.fetch(context.filename)
    if package is None:
        return HTTPBadRequest("Could not find %s" % context.filename)
    request.db.delete(package)
    return request.response


@view_config(
    context=APIResource,
    name="user",
    request_method="PUT",
    subpath=("username/*"),
    renderer="json",
    permission=NO_PERMISSION_REQUIRED,
)
@argify
def register(request, password):
    """Register a user"""
    username = request.named_subpaths["username"]
    return handle_register_request(request, username, password)


@view_config(
    context=APIResource,
    name="user",
    subpath=("password"),
    request_method="POST",
    permission="login",
)
@argify
def change_password(request, old_password, new_password):
    """Change a user's password"""
    if not request.access.verify_user(request.authenticated_userid, old_password):
        return HTTPForbidden()
    request.access.edit_user_password(request.authenticated_userid, new_password)
    return request.response



""" Views """
import logging
import traceback

from pyramid.httpexceptions import HTTPException, HTTPServerError
from pyramid.settings import asbool
from pyramid.view import view_config
from pyramid_duh import addslash

from pypicloud import __version__
from pypicloud.route import Root

LOG = logging.getLogger(__name__)


@view_config(context=Root, request_method="GET", subpath=(), renderer="base.jinja2")
@addslash
def get_index(request):
    """Render a home screen"""
    return {"version": __version__}


@view_config(route_name="health", renderer="json")
def health_endpoint(request):
    """Simple health endpoint"""
    ret = {}
    ok, ret["access"] = request.access.check_health()
    if not ok:
        request.response.status = 500
    ok, ret["cache"] = request.db.check_health()
    if not ok:
        request.response.status = 500
    ok, ret["storage"] = request.db.storage.check_health()
    if not ok:
        request.response.status = 500

    return ret


@view_config(context=Exception, renderer="json")
@view_config(context=HTTPException, renderer="json")
def format_exception(context, request):
    """
    Catch all app exceptions and render them nicely

    This will keep the status code, but will always return parseable json

    Returns
    -------
    error : str
        Identifying error key
    message : str
        Human-readable error message
    stacktrace : str, optional
        If pyramid.debug = true, also return the stacktrace to the client

    """
    message = context.message if hasattr(context, "message") else str(context)
    LOG.exception(message)
    if not request.path.startswith("/api/") and not request.path.startswith("/admin/"):
        if isinstance(context, HTTPException):
            return context
        else:
            return HTTPServerError(message)
    error = {"error": getattr(context, "error", "unknown"), "message": message}
    if asbool(request.registry.settings.get("pyramid.debug", False)):
        error["stacktrace"] = traceback.format_exc()
    request.response.status_code = getattr(context, "status_code", 500)
    return error



""" Abstract backends that are backed by simple JSON """

from .base import IAccessBackend, IMutableAccessBackend


class IJsonAccessBackend(IAccessBackend):
    """
    This backend reads the permissions from anything that can provide JSON data

    Notes
    -----
    JSON should look like this::

        {
            "users": {
                "user1": "hashed_password1",
                "user2": "hashed_password2",
                "user3": "hashed_password3",
                "user4": "hashed_password4",
                "user5": "hashed_password5",
            },
            "groups": {
                "admins": [
                "user1",
                "user2"
                ],
                "group1": [
                "user3"
                ]
            },
            "admins": [
                "user1"
            ]
            "packages": {
                "mypackage": {
                    "groups": {
                        "group1": ["read', "write"],
                        "group2": ["read"],
                        "group3": [],
                    },
                    "users": {
                        "user1": ["read", "write"],
                        "user2": ["read"],
                        "user3": [],
                        "user5": ["read"],
                    }
                }
            }
        }
    """

    _db = None

    @property
    def db(self):
        """Fetch JSON and cache it for future calls"""
        if self._db is None:
            self._db = self._get_db()
            for key in ["users", "groups", "packages", "pending_users"]:
                self._db.setdefault(key, {})
            self._db.setdefault("admins", [])
        return self._db

    def _get_db(self):
        """
        Actually fetch the remote json. This method should return an instance
        of a child class of IMutableJsonAccessDB.
        """
        raise NotImplementedError

    def _get_password_hash(self, username):
        return self.db["users"].get(username)

    def groups(self, username=None):
        if not username:
            return list(self.db["groups"].keys())
        ret_groups = []
        groups = self.db["groups"]
        for group_name, users in groups.items():
            if username in users:
                ret_groups.append(group_name)
        return ret_groups

    def group_members(self, group):
        return list(self.db["groups"].get(group, []))

    def is_admin(self, username):
        return username in self.db["admins"]

    def group_permissions(self, package):
        result = {}
        package_data = self.db["packages"].get(package, {})
        package_groups = package_data.get("groups", {})
        for group, permissions in package_groups.items():
            result[group] = permissions
        return result

    def user_permissions(self, package):
        result = {}
        package_data = self.db["packages"].get(package, {})
        package_users = package_data.get("users", {})
        for user, permissions in package_users.items():
            result[user] = permissions
        return result

    def user_package_permissions(self, username):
        packages = []
        for package_name, value in self.db["packages"].items():
            package_users = value.get("users", {})
            has_perms = username in package_users
            if has_perms:
                packages.append(
                    {"package": package_name, "permissions": package_users[username]}
                )
        return packages

    def group_package_permissions(self, group):
        packages = []
        for package_name, value in self.db["packages"].items():
            package_groups = value.get("groups", {})
            has_perms = group in package_groups.keys()
            if has_perms:
                packages.append(
                    {
                        "package": package_name,
                        "permissions": package_groups.get(group, []),
                    }
                )
        return packages

    def user_data(self, username=None):
        admins = self.db["admins"]
        if username:
            if username not in self.db["users"]:
                return None
            return {
                "username": username,
                "admin": username in admins,
                "groups": self.groups(username),
            }
        return [
            {"username": username, "admin": username in admins}
            for username in self.db["users"]
        ]


class IMutableJsonAccessBackend(IJsonAccessBackend, IMutableAccessBackend):
    """
    This backend allows you to store all user and package permissions in
    a backend that is able to store a json file

    Notes
    -----
    The format is the same as
    :class:`~pypicloud.access.base_json.IJsonAccessBackend`, but with the
    additional fields::

        {
            "pending_users": {
                "user1": "hashed_password1",
                "user2": "hashed_password2"
            },
            "allow_registration": true
        }

    """

    mutable = True

    def _save(self):
        """Save the JSON to the backend"""
        raise NotImplementedError

    def _set_password_hash(self, username, password_hash):
        self.db["users"][username] = password_hash
        self._save()

    def allow_register(self):
        return self.db.get("allow_registration", False)

    def _register(self, username, password):
        self.db["pending_users"][username] = password
        self._save()

    def approve_user(self, username):
        password = self.db["pending_users"].pop(username, None)
        if password is not None:
            self.db["users"][username] = password
        self._save()

    def delete_user(self, username):
        self.db["pending_users"].pop(username, None)
        self.db["users"].pop(username, None)

        for package_name, value in self.db["packages"].items():
            if "users" in value:
                value["users"].pop(username, None)
        for group_name, users in self.db["groups"].items():
            try:
                users.remove(username)
            except ValueError:
                pass
        self._save()

    def pending_users(self):
        return list(self.db["pending_users"].keys())

    def create_group(self, group):
        self.db["groups"][group] = []
        self._save()

    def delete_group(self, group):
        self.db["groups"].pop(group, None)
        self._save()

    def edit_user_group(self, username, group, add):
        if add:
            self.db["groups"][group].append(username)
        else:
            self.db["groups"][group].remove(username)
        self._save()

    def _init_package(self, package):
        """
        Make sure the root requested package and its child nodes exist in
        the database.
        """
        self.db["packages"].setdefault(package, {})
        self.db["packages"][package].setdefault("groups", {})
        self.db["packages"][package].setdefault("users", {})

    def edit_group_permission(self, package_name, group, perm, add):
        if perm != "read" and perm != "write":
            raise ValueError("Unrecognized permission '%s'" % perm)
        self._init_package(package_name)
        package = self.db["packages"][package_name]
        if group not in package["groups"]:
            package["groups"][group] = []
        if add:
            group_perms = package["groups"][group]
            if perm not in group_perms:
                group_perms.append(perm)
            package["groups"][group] = group_perms
        else:
            package["groups"][group].remove(perm)
            if package["groups"][group] == []:
                package["groups"].pop(group)
        self._save()

    def edit_user_permission(self, package_name, username, perm, add):
        if perm != "read" and perm != "write":
            raise ValueError("Unrecognized permission '%s'" % perm)
        self._init_package(package_name)
        package = self.db["packages"][package_name]
        if username not in package["users"]:
            package["users"][username] = []
        if add:
            user_perms = package["users"][username]
            if perm not in user_perms:
                user_perms.append(perm)
            package["users"][username] = user_perms
        else:
            try:
                package["users"][username].remove(perm)
            except ValueError:
                pass
            user_perms = package["users"][username]
            if user_perms == []:
                package["users"].pop(username)
        self._save()

    def set_user_admin(self, username, admin):
        if admin:
            self.db["admins"].append(username)
        else:
            self.db["admins"].remove(username)
        self._save()

    def set_allow_register(self, allow):
        self.db["allow_registration"] = allow
        self._save()



""" API endpoints for admin controls """
import gzip
import json
from io import BytesIO

from paste.httpheaders import CONTENT_DISPOSITION  # pylint: disable=E0611
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.response import FileIter
from pyramid.view import view_config, view_defaults
from pyramid_duh import argify

from pypicloud.route import AdminResource


@view_defaults(context=AdminResource, subpath=(), permission="admin", renderer="json")
class AdminEndpoints(object):

    """Collection of admin endpoints"""

    def __init__(self, request):
        self.request = request

    @view_config(name="rebuild")
    def rebuild_package_list(self):
        """Rebuild the package cache in the database"""
        self.request.db.reload_from_storage()
        return self.request.response

    @view_config(name="pending_users", request_method="GET")
    def get_pending_users(self):
        """Get the list of pending users"""
        return self.request.access.pending_users()

    @view_config(name="user", request_method="GET")
    def get_users(self):
        """Get the list of users"""
        return self.request.access.user_data()

    @view_config(name="user", subpath=("username/*"), request_method="GET")
    def get_user(self):
        """Get a single user"""
        username = self.request.named_subpaths["username"]
        return self.request.access.user_data(username)

    @view_config(name="user", subpath=("username/*"), request_method="PUT")
    @argify
    def create_user(self, password):
        """Create a new user"""
        username = self.request.named_subpaths["username"]
        self.request.access.register(username, password)
        self.request.access.approve_user(username)
        return self.request.response

    @view_config(name="user", subpath=("username/*"), request_method="DELETE")
    def delete_user(self):
        """Delete a user"""
        username = self.request.named_subpaths["username"]
        self.request.access.delete_user(username)
        return self.request.response

    @view_config(name="user", subpath=("username/*", "approve"), request_method="POST")
    def approve_user(self):
        """Approve a pending user"""
        username = self.request.named_subpaths["username"]
        self.request.access.approve_user(username)
        return self.request.response

    @view_config(name="user", subpath=("username/*", "admin"), request_method="POST")
    @argify
    def set_admin_status(self, admin):
        """Set a user to be or not to be an admin"""
        username = self.request.named_subpaths["username"]
        self.request.access.set_user_admin(username, admin)
        return self.request.response

    @view_config(
        name="user", subpath=("username/*", "group", "group/*"), request_method="PUT"
    )
    @view_config(
        name="user", subpath=("username/*", "group", "group/*"), request_method="DELETE"
    )
    def mutate_group_member(self):
        """Add a user to a group"""
        username = self.request.named_subpaths["username"]
        group = self.request.named_subpaths["group"]
        self.request.access.edit_user_group(
            username, group, self.request.method == "PUT"
        )
        return self.request.response

    @view_config(name="group", request_method="GET")
    def get_groups(self):
        """Get the list of groups"""
        return self.request.access.groups()

    @view_config(name="group", subpath=("group/*"), request_method="PUT")
    def create_group(self):
        """Create a group"""
        group = self.request.named_subpaths["group"]
        if group in ("everyone", "authenticated"):
            return HTTPBadRequest("'%s' is a reserved name" % group)
        self.request.access.create_group(group)
        return self.request.response

    @view_config(name="group", subpath=("group/*"), request_method="DELETE")
    def delete_group(self):
        """Delete a group"""
        group = self.request.named_subpaths["group"]
        self.request.access.delete_group(group)
        return self.request.response

    @view_config(name="user", subpath=("username/*", "permissions"))
    def get_user_permissions(self):
        """Get the package permissions for a user"""
        username = self.request.named_subpaths["username"]
        return self.request.access.user_package_permissions(username)

    @view_config(name="group", subpath=("group/*"))
    def get_group(self):
        """Get the members and package permissions for a group"""
        group = self.request.named_subpaths["group"]
        return {
            "members": self.request.access.group_members(group),
            "packages": self.request.access.group_package_permissions(group),
        }

    @view_config(name="package", subpath=("package/*"), request_method="GET")
    def get_package_permissions(self):
        """Get the user and group permissions set on a package"""
        package = self.request.named_subpaths["package"]
        user_perms = [
            {"username": key, "permissions": val}
            for key, val in self.request.access.user_permissions(package).items()
        ]
        group_perms = [
            {"group": key, "permissions": val}
            for key, val in self.request.access.group_permissions(package).items()
        ]
        return {"user": user_perms, "group": group_perms}

    @view_config(
        name="package",
        subpath=("package/*", "type/user|group/r", "name/*", "permission/read|write/r"),
        request_method="PUT",
    )
    @view_config(
        name="package",
        subpath=("package/*", "type/user|group/r", "name/*", "permission/read|write/r"),
        request_method="DELETE",
    )
    def edit_permission(self):
        """Edit user permission on a package"""
        package = self.request.named_subpaths["package"]
        name = self.request.named_subpaths["name"]
        permission = self.request.named_subpaths["permission"]
        owner_type = self.request.named_subpaths["type"]
        if owner_type == "user":
            self.request.access.edit_user_permission(
                package, name, permission, self.request.method == "PUT"
            )
        else:
            self.request.access.edit_group_permission(
                package, name, permission, self.request.method == "PUT"
            )
        return self.request.response

    @view_config(name="register", request_method="POST")
    @argify
    def toggle_allow_register(self, allow):
        """Allow or disallow user registration"""
        self.request.access.set_allow_register(allow)
        return self.request.response

    @view_config(name="token", subpath=("username/*"), request_method="GET")
    def generate_token(self):
        """Create a signup token for a user"""
        username = self.request.named_subpaths["username"]
        token = self.request.access.get_signup_token(username)
        token_url = self.request.app_url("login") + "#/?token=" + token
        return {"token": token, "token_url": token_url}

    @view_config(name="acl.json.gz", request_method="GET")
    def download_access_control(self):
        """Download the ACL data as a gzipped-json file"""
        data = self.request.access.dump()
        compressed = BytesIO()
        zipfile = gzip.GzipFile(mode="wb", fileobj=compressed)
        zipfile.write(json.dumps(data, separators=(",", ":")).encode("utf8"))
        zipfile.close()
        compressed.seek(0)

        disp = CONTENT_DISPOSITION.tuples(filename="acl.json.gz")
        self.request.response.headers.update(disp)
        self.request.response.app_iter = FileIter(compressed)
        return self.request.response



""" The access backend object base class """
import hashlib
import hmac
import time
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Tuple

from passlib.apps import LazyCryptContext
from passlib.utils import sys_bits
from pyramid.security import ALL_PERMISSIONS, Allow, Authenticated, Deny, Everyone
from pyramid.settings import aslist

from pypicloud.util import get_environ_setting

# Roughly tuned using https://bitbucket.org/ecollins/passlib/raw/default/choose_rounds.py
# For 10ms. This differs from the passlib recommendation of 350ms due to the difference in use case
DEFAULT_ROUNDS = {
    "sha512_crypt__default_rounds": 20500,
    "sha256_crypt__default_rounds": 16000,
    "pbkdf2_sha512__default_rounds": 10500,
    "pbkdf2_sha256__default_rounds": 13000,
    "bcrypt__default_rounds": 7,
    "argon2__default_rounds": 1,
}
SCHEMES = [
    "sha512_crypt",
    "sha256_crypt",
    "pbkdf2_sha512",
    "pbkdf2_sha256",
    "bcrypt",
    "argon2",
]
if sys_bits < 64:
    SCHEMES.remove("sha512_crypt")
    SCHEMES.remove("pbkdf2_sha512")


def get_pwd_context(
    preferred_hash: Optional[str] = None, rounds: Optional[int] = None
) -> LazyCryptContext:
    """Create a passlib context for hashing passwords"""
    if preferred_hash is None or preferred_hash == "sha":
        preferred_hash = "sha256_crypt" if sys_bits < 64 else "sha512_crypt"
    if preferred_hash == "pbkdf2":
        preferred_hash = "pbkdf2_sha256" if sys_bits < 64 else "pbkdf2_sha512"

    if preferred_hash not in SCHEMES:
        raise Exception(
            "Password hash %r is not in the list of supported schemes" % preferred_hash
        )

    # Put the preferred hash at the beginning of the schemes list
    schemes = list(SCHEMES)
    schemes.remove(preferred_hash)
    schemes.insert(0, preferred_hash)

    # Override the default rounds of the preferred hash, if provided
    default_rounds = dict(DEFAULT_ROUNDS)
    if rounds is not None:
        default_rounds[preferred_hash + "__default_rounds"] = rounds

    return LazyCryptContext(schemes=schemes, default=schemes[0], **default_rounds)


def group_to_principal(group: str) -> str:
    """Convert a group to its corresponding principal"""
    if group in (Everyone, Authenticated) or group.startswith("group:"):
        return group
    elif group == "everyone":
        return Everyone
    elif group == "authenticated":
        return Authenticated
    else:
        return "group:" + group


def groups_to_principals(groups: List[str]) -> List[str]:
    """Convert a list of groups to a list of principals"""
    return [group_to_principal(g) for g in groups]


ONE_WEEK = 60 * 60 * 24 * 7


class IAccessBackend(object):

    """Base class for retrieving user and package permission data"""

    mutable = False
    ROOT_ACL = [
        (Allow, Authenticated, "login"),
        (Allow, "admin", ALL_PERMISSIONS),
        (Deny, Everyone, ALL_PERMISSIONS),
    ]

    def __init__(
        self,
        request=None,
        default_read=None,
        default_write=None,
        disallow_fallback=(),
        cache_update=None,
        pwd_context=None,
        token_expiration=ONE_WEEK,
        signing_key=None,
    ):
        self.request = request
        self.default_read = default_read
        self.default_write = default_write
        self.disallow_fallback = disallow_fallback
        self.cache_update = cache_update
        self.pwd_context = pwd_context
        self.token_expiration = token_expiration
        self.signing_key = signing_key

    @classmethod
    def configure(cls, settings) -> Dict[str, Any]:
        """Configure the access backend with app settings"""
        rounds = settings.get("auth.rounds")
        scheme = settings.get("auth.scheme")
        return {
            "default_read": aslist(
                settings.get("pypi.default_read", ["authenticated"])
            ),
            "default_write": aslist(settings.get("pypi.default_write", [])),
            "disallow_fallback": aslist(settings.get("pypi.disallow_fallback", [])),
            "cache_update": aslist(
                settings.get("pypi.cache_update", ["authenticated"])
            ),
            "pwd_context": get_pwd_context(scheme, rounds),
            "token_expiration": int(settings.get("auth.token_expire", ONE_WEEK)),
            "signing_key": get_environ_setting(settings, "auth.signing_key"),
        }

    @classmethod
    def postfork(cls, **kwargs):
        """This method will be called after uWSGI forks"""

    def allowed_permissions(self, package: str) -> Dict[str, Tuple[str, ...]]:
        """
        Get all allowed permissions for all principals on a package

        Returns
        -------
        perms : dict
            Mapping of principal to tuple of permissions

        """
        all_perms = {}
        for user, perms in self.user_permissions(package).items():
            all_perms["user:" + user] = tuple(perms)

        for group, perms in self.group_permissions(package).items():
            all_perms[group_to_principal(group)] = tuple(perms)

        # If there are no group or user specifications for the package, use the
        # default
        if not all_perms:
            for principal in groups_to_principals(self.default_read):
                all_perms[principal] = ("read",)
            for principal in groups_to_principals(self.default_write):
                if principal in all_perms:
                    all_perms[principal] += ("write",)
                else:
                    all_perms[principal] = ("write",)

        # add fallback permissions
        if package not in self.disallow_fallback:
            for principal in all_perms:
                all_perms[principal] += ("fallback",)
        return all_perms

    def get_acl(self, package: str) -> List[Tuple[str, str, str]]:
        """Construct an ACL for a package"""
        acl = []
        permissions = self.allowed_permissions(package)
        for principal, perms in permissions.items():
            for perm in perms:
                acl.append((Allow, principal, perm))
        return acl

    def has_permission(self, package: str, perm: str) -> bool:
        """Check if this user has a permission for a package"""
        current_userid = self.request.authenticated_userid
        if current_userid is not None and self.is_admin(current_userid):
            return True

        perms = self.allowed_permissions(package)
        for principal in self.request.effective_principals:
            if perm in perms.get(principal, []):
                return True
        return False

    def user_principals(self, username: str) -> List[str]:
        """
        Get a list of principals for a user

        Parameters
        ----------
        username : str

        Returns
        -------
        principals : list

        """
        principals = ["user:" + username, Everyone, Authenticated]
        if self.is_admin(username):
            principals.append("admin")
        for group in self.groups(username):
            principals.append("group:" + group)
        return principals

    def in_group(self, username: str, group: str) -> bool:
        """
        Find out if a user is in a group

        Parameters
        ----------
        username : str
            Name of user. May be None for the anonymous user.
        group : str
            Name of the group. Supports 'everyone', 'authenticated', and
            'admin'.

        Returns
        -------
        member : bool

        """
        if group in ("everyone", Everyone):
            return True
        elif username is None:
            return False
        elif group in ("authenticated", Authenticated):
            return True
        elif group == "admin" and self.is_admin(username):
            return True
        else:
            return group in self.groups(username)

    def in_any_group(self, username: str, groups: List[str]) -> bool:
        """
        Find out if a user is in any of a set of groups

        Parameters
        ----------
        username : str
            Name of user. May be None for the anonymous user.
        groups : list
            list of group names. Supports 'everyone', 'authenticated', and
            'admin'.

        Returns
        -------
        member : bool

        """
        return any((self.in_group(username, group) for group in groups))

    def can_update_cache(self) -> bool:
        """
        Return True if the user has permissions to update the pypi cache
        """
        return self.in_any_group(self.request.authenticated_userid, self.cache_update)

    def need_admin(self) -> bool:
        """
        Find out if there are any admin users

        This should only be overridden by mutable backends

        Returns
        -------
        need_admin : bool
            True if no admin user exists and the backend is mutable, False
            otherwise

        """
        return False

    def allow_register(self) -> bool:
        """
        Check if the backend allows registration

        This should only be overridden by mutable backends

        Returns
        -------
        allow : bool

        """
        return False

    def allow_register_token(self) -> bool:
        """
        Check if the backend allows registration via tokens

        This should only be overridden by mutable backends

        Returns
        -------
        allow : bool

        """
        return False

    def verify_user(self, username: str, password: str) -> bool:
        """
        Check the login credentials of a user

        For Mutable backends, pending users should fail to verify

        Parameters
        ----------
        username : str
        password : str

        Returns
        -------
        valid : bool
            True if user credentials are valid, false otherwise

        """
        stored_pw = self._get_password_hash(username)
        if self.mutable:
            # if a user is pending, user_data will be None
            user_data = self.user_data(username)
            if user_data is None:
                return False
        return bool(stored_pw and self.pwd_context.verify(password, stored_pw))

    def _get_password_hash(self, username: str) -> str:
        """Get the stored password hash for a user"""
        raise NotImplementedError

    def groups(self, username: Optional[str] = None) -> List[str]:
        """
        Get a list of all groups

        If a username is specified, get all groups that the user belongs to

        Parameters
        ----------
        username : str, optional

        Returns
        -------
        groups : list
            List of group names

        """
        raise NotImplementedError

    def group_members(self, group: str) -> List[str]:
        """
        Get a list of users that belong to a group

        Parameters
        ----------
        group : str

        Returns
        -------
        users : list
            List of user names

        """
        raise NotImplementedError

    def is_admin(self, username: str) -> bool:
        """
        Check if the user is an admin

        Parameters
        ----------
        username : str

        Returns
        -------
        is_admin : bool

        """
        raise NotImplementedError

    def group_permissions(self, package: str) -> Dict[str, List[str]]:
        """
        Get a mapping of all groups to their permissions on a package

        Parameters
        ----------
        package : str
            The name of a python package


        Returns
        -------
        permissions : dict
            mapping of group name to a list of permissions
            (which can contain 'read' and/or 'write')

        """
        raise NotImplementedError

    def user_permissions(self, package: str) -> Dict[str, List[str]]:
        """
        Get a mapping of all users to their permissions for a package

        Parameters
        ----------
        package : str
            The name of a python package

        Returns
        -------
        permissions : dict
            Mapping of username to a list of permissions (which can contain
            'read' and/or 'write')

        """
        raise NotImplementedError

    def user_package_permissions(self, username: str) -> List[Dict[str, List[str]]]:
        """
        Get a list of all packages that a user has permissions on

        Parameters
        ----------
        username : str

        Returns
        -------
        packages : list
            List of dicts. Each dict contains 'package' (str) and 'permissions'
            (list)

        """
        raise NotImplementedError

    def group_package_permissions(self, group: str) -> List[Dict[str, List[str]]]:
        """
        Get a list of all packages that a group has permissions on

        Parameters
        ----------
        group : str

        Returns
        -------
        packages : list
            List of dicts. Each dict contains 'package' (str) and 'permissions'
            (list)

        """
        raise NotImplementedError

    def user_data(self, username=None):
        """
        Get a list of all users or data for a single user

        For Mutable backends, this MUST exclude all pending users

        Returns
        -------
        users : list
            Each user is a dict with a 'username' str, and 'admin' bool
        user : dict
            If a username is passed in, instead return one user with the fields
            above plus a 'groups' list.

        """
        raise NotImplementedError

    def check_health(self) -> Tuple[bool, str]:
        """
        Check the health of the access backend

        Returns
        -------
        (healthy, status) : (bool, str)
            Tuple that describes the health status and provides an optional
            status message

        """
        return True, ""

    def dump(self) -> Dict[str, Any]:
        """
        Dump all of the access control data to a universal format

        Returns
        -------
        data : dict

        """
        from pypicloud import __version__

        data = {}  # type: Dict[str, Any]
        data["allow_register"] = self.allow_register()
        data["version"] = __version__

        groups = self.groups()
        users = self.user_data()
        for user in users:
            user["password"] = self._get_password_hash(user["username"])

        data["groups"] = {}
        packages = {
            "users": defaultdict(dict),
            "groups": defaultdict(dict),
        }  # type: Dict[str, Any]
        for group in groups:
            data["groups"][group] = self.group_members(group)
            perms = self.group_package_permissions(group)
            for perm in perms:
                package = perm["package"]
                packages["groups"][package][group] = perm["permissions"]

        for user in users:
            username = user["username"]
            perms = self.user_package_permissions(username)
            for perm in perms:
                package = perm["package"]
                packages["users"][package][username] = perm["permissions"]

        # Convert the defaultdict to a dict for easy serialization
        packages["users"] = dict(packages["users"])
        packages["groups"] = dict(packages["groups"])
        data["users"] = users
        data["packages"] = packages
        return data

    def load(self, data):
        """
        Idempotently load universal access control data.

        By default, this does nothing on immutable backends. Backends may
        override this method to provide an implementation.

        This method works by default on mutable backends with no override
        necessary.

        """
        raise TypeError(
            "Access backend '%s' is not mutable and has no "
            "'load' implementation" % self.__class__.__name__
        )


class IMutableAccessBackend(IAccessBackend):

    """
    Base class for access backends that can change user/group permissions

    """

    mutable = True

    def need_admin(self) -> bool:
        for user in self.user_data():
            if user["admin"]:
                return False
        return True

    def get_signup_token(self, username: str) -> str:
        """
        Create a signup token

        Parameters
        ----------
        username : str
            The username to be created when this token is consumed

        Returns
        -------
        token : str

        """
        msg, signature = self._hmac(username, time.time())
        return msg + ":" + signature

    def _hmac(self, username: str, timestamp: float) -> Tuple[str, str]:
        """HMAC a username/expiration combo"""
        if self.signing_key is None:
            raise RuntimeError("auth.signing_key is not set!")
        msg = "%s:%d" % (username, timestamp)
        return (
            msg,
            hmac.new(
                self.signing_key.encode("utf8"), msg.encode("utf8"), hashlib.sha256
            ).hexdigest(),
        )

    def validate_signup_token(self, token: str) -> Optional[str]:
        """
        Validate a signup token

        Parameters
        ----------
        token : str

        Returns
        -------
        username : str or None
            This will be None if the validation fails

        """
        if self.signing_key is None:
            return None
        pieces = token.split(":")
        signature = pieces.pop()
        username = pieces[0]
        issued = int(pieces[1])
        if issued + self.token_expiration < time.time():
            return None
        _, expected = self._hmac(username, issued)
        if hasattr(hmac, "compare_digest"):
            if not hmac.compare_digest(
                signature.encode("utf-8"), expected.encode("utf-8")
            ):
                return None
        else:
            if signature != expected:
                return None
        return username

    def allow_register(self):
        raise NotImplementedError

    def allow_register_token(self):
        return self.signing_key is not None

    def set_allow_register(self, allow: bool) -> None:
        """
        Allow or disallow user registration

        Parameters
        ----------
        allow : bool

        """
        raise NotImplementedError

    def register(self, username: str, password: str) -> None:
        """
        Register a new user

        The new user should be marked as pending admin approval

        Parameters
        ----------
        username : str
        password : str
            This should be the plaintext password

        """
        self._register(username, self.pwd_context.hash(password))

    def _register(self, username: str, password: str) -> None:
        """
        Register a new user

        The new user should be marked as pending admin approval

        Parameters
        ----------
        username : str
        password : str
            This will be the hash of the password

        """
        raise NotImplementedError

    def pending_users(self) -> List[str]:
        """
        Retrieve a list of all users pending admin approval

        Returns
        -------
        users : list
            List of usernames

        """
        raise NotImplementedError

    def approve_user(self, username: str) -> None:
        """
        Mark a user as approved by the admin

        Parameters
        ----------
        username : str

        """
        raise NotImplementedError

    def edit_user_password(self, username: str, password: str) -> None:
        """
        Change a user's password

        Parameters
        ----------
        username : str
        password : str

        """
        self._set_password_hash(username, self.pwd_context.hash(password))

    def _set_password_hash(self, username: str, password_hash: str) -> None:
        """
        Change a user's password

        Parameters
        ----------
        username : str
        password_hash : str
            The hashed password to store

        """
        raise NotImplementedError

    def delete_user(self, username: str) -> None:
        """
        Delete a user

        Parameters
        ----------
        username : str

        """
        raise NotImplementedError

    def set_user_admin(self, username: str, admin: bool) -> None:
        """
        Grant or revoke admin permissions for a user

        Parameters
        ----------
        username : str
        admin : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def edit_user_group(self, username: str, group: str, add: bool) -> None:
        """
        Add or remove a user to/from a group

        Parameters
        ----------
        username : str
        group : str
        add : bool
            If True, add to group. If False, remove.

        """
        raise NotImplementedError

    def create_group(self, group: str) -> None:
        """
        Create a new group

        Parameters
        ----------
        group : str

        """
        raise NotImplementedError

    def delete_group(self, group: str) -> None:
        """
        Delete a group

        Parameters
        ----------
        group : str

        """
        raise NotImplementedError

    def edit_user_permission(
        self, package: str, username: str, perm: Set[str], add: bool
    ) -> None:
        """
        Grant or revoke a permission for a user on a package

        Parameters
        ----------
        package : str
        username : str
        perm : {'read', 'write'}
        add : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def edit_group_permission(
        self, package: str, group: str, perm: Set[str], add: bool
    ) -> None:
        """
        Grant or revoke a permission for a group on a package

        Parameters
        ----------
        package : str
        group : str
        perm : {'read', 'write'}
        add : bool
            If True, grant permissions. If False, revoke.

        """
        raise NotImplementedError

    def dump(self):
        data = super(IMutableAccessBackend, self).dump()
        pending_users = []
        for username in self.pending_users():  # pylint: disable=E1101
            password = self._get_password_hash(username)
            pending_users.append({"username": username, "password": password})
        data["pending_users"] = pending_users
        return data

    def load(self, data):
        pending_users = set(self.pending_users())

        def user_exists(username):
            """Helper function that checks if a user already exists"""
            return username in pending_users or self.user_data(username) is not None

        for user in data["users"]:
            if not user_exists(user["username"]):
                self._register(user["username"], user["password"])
                self.approve_user(user["username"])
            self.set_user_admin(user["username"], user.get("admin", False))

        for group, members in data["groups"].items():
            if not self.group_members(group):
                self.create_group(group)
            current_members = self.group_members(group)
            add_members = set(members) - set(current_members)
            for member in add_members:
                self.edit_user_group(member, group, True)

        for user in data.get("pending_users", []):
            if not user_exists(user["username"]):
                self._register(user["username"], user["password"])

        for package, groups in data["packages"]["groups"].items():
            for group, permissions in groups.items():
                for perm in permissions:
                    self.edit_group_permission(package, group, perm, True)

        for package, users in data["packages"]["users"].items():
            for user, permissions in users.items():
                for perm in permissions:
                    self.edit_user_permission(package, user, perm, True)

        self.set_allow_register(data["allow_register"])



""" Setup file """
import os
import re

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, "README.rst")).read()
CHANGES = open(os.path.join(HERE, "CHANGES.rst")).read()
# Remove custom RST extensions for pypi
CHANGES = re.sub(r"\(\s*:(issue|pr|sha):.*?\)", "", CHANGES)
CHANGES = re.sub(r":ref:`(.*?) <.*>`", r"\1", CHANGES)

REQUIREMENTS_TEST = open(os.path.join(HERE, "requirements_test.txt")).readlines()
REQUIREMENTS = [
    "boto3>=1.7.0",
    # beaker needs this
    "cryptography",
    "distlib",
    "paste",
    "passlib>=1.7",
    "pyramid>=2",
    "pyramid_beaker",
    "pyramid_duh",
    "pyramid_jinja2",
    "pyramid_rpc",
    "pyramid_tm",
    "requests",
    "transaction",
    "zope.sqlalchemy",
]

EXTRAS = {
    "ldap": ["python-ldap"],
    "dynamo": ["flywheel >= 0.2.0"],
    "redis": ["redis"],
    "gcs": ["google-cloud-storage>=1.10.0"],
    "azure-blob": ["azure-storage-blob>=12.3.1"],
}

EXTRAS["all_plugins"] = sum(EXTRAS.values(), [])

EXTRAS["server"] = ["waitress"]


if __name__ == "__main__":
    setup(
        name="pypicloud",
        version="1.2.3",
        description="Private PyPI backed by S3",
        long_description=README + "\n\n" + CHANGES,
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Development Status :: 4 - Beta",
            "Framework :: Pyramid",
            "Intended Audience :: System Administrators",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: System :: Systems Administration",
        ],
        license="MIT",
        author="Steven Arcangeli",
        author_email="stevearc@stevearc.com",
        url="http://pypicloud.readthedocs.org/",
        keywords="pypi s3 cheeseshop package",
        platforms="any",
        zip_safe=False,
        python_requires=">=3.5",
        include_package_data=True,
        packages=find_packages(exclude=("tests",)),
        entry_points={
            "console_scripts": [
                "pypicloud-gen-password = pypicloud.scripts:gen_password",
                "pypicloud-make-config = pypicloud.scripts:make_config",
                "ppc-gen-password = pypicloud.scripts:gen_password",
                "ppc-make-config = pypicloud.scripts:make_config",
                "ppc-migrate = pypicloud.scripts:migrate_packages",
                "ppc-export = pypicloud.scripts:export_access",
                "ppc-import = pypicloud.scripts:import_access",
                "ppc-create-s3-sync = pypicloud.lambda_scripts:create_sync_scripts",
                "ppc-build-lambda-bundle = pypicloud.lambda_scripts:build_lambda_bundle",
            ],
            "paste.app_factory": ["main = pypicloud:main"],
        },
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + REQUIREMENTS_TEST,
        test_suite="tests",
        extras_require=EXTRAS,
    )



""" Caching database implementations """
from functools import partial

from pyramid.path import DottedNameResolver

from .base import ICache
from .redis_cache import RedisCache
from .sql import SQLCache


def get_cache_impl(settings):
    """Get the cache class from settings"""
    resolver = DottedNameResolver(__name__)
    dotted_cache = settings.get("pypi.db", "sql")
    if dotted_cache == "sql":
        dotted_cache = "pypicloud.cache.SQLCache"
    elif dotted_cache == "redis":
        dotted_cache = "pypicloud.cache.RedisCache"
    elif dotted_cache == "dynamo":
        dotted_cache = "pypicloud.cache.dynamo.DynamoCache"
    return resolver.resolve(dotted_cache)


def includeme(config):
    """Get and configure the cache db wrapper"""
    settings = config.get_settings()
    cache_impl = get_cache_impl(settings)
    kwargs = cache_impl.configure(settings)
    cache = cache_impl(**kwargs)
    cache.reload_if_needed()
    config.add_request_method(partial(cache_impl, **kwargs), name="db", reify=True)
    config.add_postfork_hook(partial(cache_impl.postfork, **kwargs))
    return cache_impl



""" Store package data in DynamoDB """
import logging
from collections import defaultdict
from datetime import datetime

from dynamo3 import DynamoDBConnection
from pkg_resources import parse_version
from pyramid.settings import asbool, aslist

from pypicloud.dateutil import UTC, utcnow
from pypicloud.models import Package

from .base import ICache

try:
    from flywheel import Engine, Field, GlobalIndex, Model, __version__

    if parse_version(__version__) < parse_version("0.2.0"):  # pragma: no cover
        raise ValueError("Pypicloud requires flywheel>=0.2.0")
except ImportError as e:  # pragma: no cover
    raise ImportError(
        "You must 'pip install flywheel' before using " "DynamoDB as the cache database"
    ) from e

LOG = logging.getLogger(__name__)


class DynamoPackage(Package, Model):

    """Python package stored in DynamoDB"""

    __metadata__ = {"global_indexes": [GlobalIndex("name-index", "name")]}
    filename = Field(hash_key=True)
    name = Field()
    version = Field()
    last_modified = Field(data_type=datetime)
    summary = Field()
    data = Field(data_type=dict)

    def __init__(self, *args, **kwargs):
        super(DynamoPackage, self).__init__(*args, **kwargs)
        # DynamoDB doesn't play nice with empty strings.
        if not self.summary:
            self.summary = None


class PackageSummary(Model):

    """Aggregate data about packages"""

    name = Field(hash_key=True)
    summary = Field()
    last_modified = Field(data_type=datetime)

    def __init__(self, package):
        super(PackageSummary, self).__init__(package.name)
        self.last_modified = package.last_modified.replace(tzinfo=UTC)
        self.summary = package.summary


class DynamoCache(ICache):

    """Caching database that uses DynamoDB"""

    def __init__(self, request=None, engine=None, graceful_reload=False, **kwargs):
        super(DynamoCache, self).__init__(request, **kwargs)
        self.engine = engine
        self.graceful_reload = graceful_reload

    def new_package(self, *args, **kwargs):
        return DynamoPackage(*args, **kwargs)

    @classmethod
    def configure(cls, settings):
        kwargs = super(DynamoCache, cls).configure(settings)

        access_key = settings.get("db.aws_access_key_id")
        secret_key = settings.get("db.aws_secret_access_key")
        region = settings.get("db.region_name")
        host = settings.get("db.host")
        port = int(settings.get("db.port", 8000))
        secure = asbool(settings.get("db.secure", False))
        namespace = settings.get("db.namespace", ())
        graceful_reload = asbool(settings.get("db.graceful_reload", False))

        tablenames = aslist(settings.get("db.tablenames", []))
        if tablenames:
            if len(tablenames) != 2:
                raise ValueError("db.tablenames must be a 2-element list")
            DynamoPackage.meta_.name = tablenames[0]
            PackageSummary.meta_.name = tablenames[1]

        if host is not None:
            connection = DynamoDBConnection.connect(
                region,
                host=host,
                port=port,
                is_secure=secure,
                access_key=access_key,
                secret_key=secret_key,
            )
        elif region is not None:
            connection = DynamoDBConnection.connect(
                region, access_key=access_key, secret_key=secret_key
            )
        else:
            raise ValueError("Must specify either db.region_name or db.host!")
        kwargs["engine"] = engine = Engine(namespace=namespace, dynamo=connection)
        kwargs["graceful_reload"] = graceful_reload

        engine.register(DynamoPackage, PackageSummary)
        LOG.info("Checking if DynamoDB tables exist")
        engine.create_schema()
        return kwargs

    def fetch(self, filename):
        return self.engine.get(DynamoPackage, filename=filename)

    def all(self, name):
        return sorted(self.engine.query(DynamoPackage).filter(name=name), reverse=True)

    def distinct(self):
        names = set()
        for summary in self.engine.scan(PackageSummary):
            names.add(summary.name)
        return sorted(names)

    def summary(self):
        summaries = sorted(self.engine.scan(PackageSummary), key=lambda s: s.name)
        return [s.__json__() for s in summaries]

    def clear(self, package):
        self.engine.delete(package)
        self._maybe_delete_summary(package.name)

    def _maybe_delete_summary(self, package_name):
        """Check for any package with the name. Delete summary if 0"""
        remaining = (
            self.engine(DynamoPackage)
            .filter(DynamoPackage.name == package_name)
            .scan_limit(1)
            .count()
        )
        if remaining == 0:
            LOG.info("Removing package summary %s", package_name)
            self.engine.delete_key(PackageSummary, name=package_name)

    def clear_all(self):
        # We're replacing the schema, so make sure we save and restore the
        # current table/index throughput
        throughput = {}
        for model in (DynamoPackage, PackageSummary):
            tablename = model.meta_.ddb_tablename(self.engine.namespace)
            desc = self.engine.dynamo.describe_table(tablename)
            tablename = model.meta_.ddb_tablename()
            throughput[tablename] = {
                "read": desc.throughput.read,
                "write": desc.throughput.write,
            }
            for index in desc.global_indexes:
                throughput[tablename][index.name] = {
                    "read": index.throughput.read,
                    "write": index.throughput.write,
                }

        self.engine.delete_schema()
        self.engine.create_schema(throughput=throughput)

    def save(self, package):
        summary = PackageSummary(package)
        self.engine.save([package, summary], overwrite=True)

    def reload_from_storage(self, clear=True):
        if not self.graceful_reload:
            return super(DynamoCache, self).reload_from_storage(clear)
        LOG.info("Rebuilding cache from storage")
        # Log start time
        start = utcnow()
        # Fetch packages from storage s1
        s1 = set(self.storage.list(self.new_package))
        # Fetch cache packages c1
        c1 = set(self.engine.scan(DynamoPackage))
        # Add missing packages to cache (s1 - c1)
        missing = s1 - c1
        if missing:
            LOG.info("Adding %d missing packages to cache", len(missing))
            self.engine.save(missing)
        # Delete extra packages from cache (c1 - s1) when last_modified < start
        # The time filter helps us avoid deleting packages that were
        # concurrently uploaded.
        extra1 = [p for p in (c1 - s1) if p.last_modified < start]
        if extra1:
            LOG.info("Removing %d extra packages from cache", len(extra1))
            self.engine.delete(extra1)

        # If any packages were concurrently deleted during the cache rebuild,
        # we can detect them by polling storage again and looking for any
        # packages that were present in s1 and are missing from s2
        s2 = set(self.storage.list(self.new_package))
        # Delete extra packages from cache (s1 - s2)
        extra2 = s1 - s2
        if extra2:
            LOG.info(
                "Removing %d packages from cache that were concurrently "
                "deleted during rebuild",
                len(extra2),
            )
            self.engine.delete(extra2)
            # Remove these concurrently-deleted files from the list of packages
            # that were missing from the cache. Don't want to use those to
            # update the summaries below.
            missing -= extra2

        # Update the PackageSummary for added packages
        packages_by_name = defaultdict(list)
        for package in missing:
            # Set the tz here so we can compare against the PackageSummary
            package.last_modified = package.last_modified.replace(tzinfo=UTC)
            packages_by_name[package.name].append(package)
        summaries = self.engine.get(PackageSummary, packages_by_name.keys())
        summaries_by_name = {}
        for summary in summaries:
            summaries_by_name[summary.name] = summary
        for name, packages in packages_by_name.items():
            if name in summaries_by_name:
                summary = summaries_by_name[name]
            else:
                summary = PackageSummary(packages[0])
                summaries.append(summary)
            for package in packages:
                if package.last_modified > summary.last_modified:
                    summary.last_modified = package.last_modified
                    summary.summary = package.summary
        if summaries:
            LOG.info("Updating %d package summaries", len(summaries))
            self.engine.save(summaries, overwrite=True)

        # Remove the PackageSummary for deleted packages
        removed = set()
        for package in extra1:
            removed.add(package.name)
        for package in extra2:
            removed.add(package.name)
        for name in removed:
            self._maybe_delete_summary(name)

    def check_health(self):
        try:
            self.engine.scan(PackageSummary).first()
        except Exception as e:
            return False, str(e)
        else:
            return True, ""



""" Store packages in S3 """
import logging
import posixpath
from datetime import timedelta
from urllib.parse import quote, urlparse

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from botocore.signers import CloudFrontSigner
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from pyramid.settings import asbool, falsey
from pyramid_duh.settings import asdict

from pypicloud.dateutil import utcnow
from pypicloud.models import Package
from pypicloud.util import get_settings, normalize_metadata, parse_filename

from .object_store import ObjectStoreStorage

LOG = logging.getLogger(__name__)


class S3Storage(ObjectStoreStorage):

    """Storage backend that uses S3"""

    test = False

    def __init__(self, request=None, bucket=None, **kwargs):
        super(S3Storage, self).__init__(request=request, **kwargs)
        self.bucket = bucket

    @classmethod
    def _subclass_specific_config(cls, settings, common_config):
        sse = settings.get("storage.server_side_encryption")
        if sse not in [None, "AES256", "aws:kms"]:
            LOG.warning(
                "Unrecognized value %r for 'storage.sse'. See "
                "https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Object.put "
                "for more details",
                sse,
            )

        bucket_name = settings.get("storage.bucket")
        if bucket_name is None:
            raise ValueError("You must specify the 'storage.bucket'")

        return {"sse": sse, "bucket": cls.get_bucket(bucket_name, settings)}

    @classmethod
    def get_bucket(cls, bucket_name, settings):
        config_settings = get_settings(
            settings,
            "storage.",
            region_name=str,
            signature_version=str,
            user_agent=str,
            user_agent_extra=str,
            connect_timeout=int,
            read_timeout=int,
            parameter_validation=asbool,
            max_pool_connections=int,
            proxies=asdict,
        )
        config_settings["s3"] = get_settings(
            settings,
            "storage.",
            use_accelerate_endpoint=asbool,
            payload_signing_enabled=asbool,
            addressing_style=str,
            signature_version=str,
        )
        config = Config(**config_settings)

        def verify_value(val):
            """Verify can be a boolean (False) or a string"""
            s = str(val).strip().lower()
            if s in falsey:
                return False
            else:
                return str(val)

        s3conn = boto3.resource(
            "s3",
            config=config,
            **get_settings(
                settings,
                "storage.",
                region_name=str,
                api_version=str,
                use_ssl=asbool,
                verify=verify_value,
                endpoint_url=str,
                aws_access_key_id=str,
                aws_secret_access_key=str,
                aws_session_token=str,
            )
        )

        bucket = s3conn.Bucket(bucket_name)
        try:
            head = s3conn.meta.client.head_bucket(Bucket=bucket_name)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                LOG.info("Creating S3 bucket %s", bucket_name)

                if config.region_name:
                    location = {"LocationConstraint": config.region_name}
                    bucket.create(CreateBucketConfiguration=location)
                else:
                    bucket.create()

                bucket.wait_until_exists()
            else:
                if e.response["Error"]["Code"] == "301":
                    LOG.error(
                        "Bucket found in different region. Check that "
                        "the S3 bucket specified in 'storage.bucket' is "
                        "in 'storage.region_name'"
                    )
                raise
        return bucket

    @classmethod
    def package_from_object(cls, obj, factory):
        """Create a package from a S3 object"""
        filename = posixpath.basename(obj.key)
        name = obj.metadata.get("name")
        version = obj.metadata.get("version")
        metadata = Package.read_metadata(obj.metadata)
        # We used to not store metadata. This is for backwards
        # compatibility
        if name is None or version is None:
            try:
                name, version = parse_filename(filename)
            except ValueError:
                LOG.warning("S3 file %s has no package name", obj.key)
                return None

        return factory(
            name, version, filename, obj.last_modified, path=obj.key, **metadata
        )

    def list(self, factory=Package):
        keys = self.bucket.objects.filter(Prefix=self.bucket_prefix)
        for summary in keys:
            # ObjectSummary has no metadata, so we have to fetch it.
            obj = summary.Object()
            pkg = self.package_from_object(obj, factory)
            if pkg is not None:
                yield pkg

    def _generate_url(self, package):
        """Generate a signed url to the S3 file"""
        if self.public_url:
            if self.region_name:
                return "https://s3.{0}.amazonaws.com/{1}/{2}".format(
                    self.region_name, self.bucket.name, self.get_path(package)
                )
            else:
                if "." in self.bucket.name:
                    self._log_region_warning()
                return "https://{0}.s3.amazonaws.com/{1}".format(
                    self.bucket.name, self.get_path(package)
                )
        url = self.bucket.meta.client.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.bucket.name, "Key": self.get_path(package)},
            ExpiresIn=self.expire_after,
        )
        # There is a special case if your bucket has a '.' in the name. The
        # generated URL will return a 301 and the pip downloads will fail.
        # If you provide a region_name, boto should correctly generate a url in
        # the form of `s3.<region>.amazonaws.com`
        # See https://github.com/stevearc/pypicloud/issues/145
        if "." in self.bucket.name:
            pieces = urlparse(url)
            if pieces.netloc == "s3.amazonaws.com" and self.region_name is None:
                self._log_region_warning()
        return url

    def _log_region_warning(self):
        """Spit out a warning about including region_name"""
        LOG.warning(
            "Your signed S3 urls may not work! "
            "Try adding the bucket region to the config with "
            "'storage.region_name = <region>' or using a bucket "
            "without any dots ('.') in the name."
        )

    def upload(self, package, datastream):
        key = self.bucket.Object(self.get_path(package))
        kwargs = {}
        if self.sse is not None:
            kwargs["ServerSideEncryption"] = self.sse
        if self.object_acl:
            kwargs["ACL"] = self.object_acl
        if self.storage_class is not None:
            kwargs["StorageClass"] = self.storage_class
        metadata = package.get_metadata()
        metadata["name"] = package.name
        metadata["version"] = package.version
        metadata = normalize_metadata(metadata)
        key.put(Metadata=metadata, Body=datastream, **kwargs)

    def delete(self, package):
        self.bucket.delete_objects(
            Delete={"Objects": [{"Key": self.get_path(package)}]}
        )

    def check_health(self):
        try:
            self.bucket.meta.client.head_bucket(Bucket=self.bucket.name)
        except ClientError as e:
            return False, str(e)
        else:
            return True, ""


class CloudFrontS3Storage(S3Storage):

    """Storage backend that uses S3 and CloudFront"""

    def __init__(
        self, request=None, domain=None, crypto_pk=None, key_id=None, **kwargs
    ):
        super(CloudFrontS3Storage, self).__init__(request, **kwargs)
        self.domain = domain
        self.crypto_pk = crypto_pk
        self.key_id = key_id

        self.cf_signer = None
        if key_id is not None:
            self.cf_signer = CloudFrontSigner(self.key_id, self._rsa_signer)

        self.client = boto3.client("cloudfront")

    @classmethod
    def configure(cls, settings):
        kwargs = super(CloudFrontS3Storage, cls).configure(settings)
        kwargs["domain"] = settings["storage.cloud_front_domain"]
        kwargs["key_id"] = settings.get("storage.cloud_front_key_id")
        private_key = settings.get("storage.cloud_front_key_string")
        if private_key is None:
            key_file = settings.get("storage.cloud_front_key_file")
            if key_file:
                with open(key_file, "rb") as ifile:
                    private_key = ifile.read()
        else:
            private_key = private_key.encode("utf-8")
        crypto_pk = serialization.load_pem_private_key(
            private_key, password=None, backend=default_backend()
        )
        kwargs["crypto_pk"] = crypto_pk

        return kwargs

    def _rsa_signer(self, message):
        """Generate a RSA signature for a message"""
        return self.crypto_pk.sign(message, padding.PKCS1v15(), hashes.SHA1())

    def _generate_url(self, package):
        """Get the fully-qualified CloudFront path for a package"""
        path = self.get_path(package)
        url = self.domain + "/" + quote(path)

        # No key id, no signer, so we don't have to sign the URL
        if self.cf_signer is None:
            return url

        # To sign with a canned policy:
        expires = utcnow() + timedelta(seconds=self.expire_after)
        return self.cf_signer.generate_presigned_url(url, date_less_than=expires)
